
/*
 *@@sourcefile bs_string.cpp:
 *      BSString implementation.
 *
 *@@header "base\bs_string.h"
 */

/*
 *      This file Copyright (C) 1999-2001 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#define OS2EMX_PLAIN_CHAR
    // this is needed for "os2emx.h"; if this is defined,
    // emx will define PSZ as _signed_ char, otherwise
    // as unsigned char

#include <os2.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "setup.h"

#include "helpers\stringh.h"
#include "helpers\xstring.h"

// base includes
#include "base\bs_base.h"
#include "base\bs_string.h"

#pragma hdrstop

// #define dprintf printf

DEFINE_CLASS(BSString, BSRoot);

/* ******************************************************************
 *
 *  BSString implementation
 *
 ********************************************************************/

// define npos; this must be in the implementation, or we'll
// get duplicate symbols from the linker
const size_type BSString::npos = (size_type)(-1);

/*
 *@@ Init:
 *      protected helper method to free
 *      allocated resources. This is used
 *      by all constructors.
 *
 *      This is an internal method. Do not
 *      call this from user code.
 */

void BSString::Init()
{
    xstrInit(&_str, 0);
    _pShiftTable = 0;
    _fRepeat = FALSE;
}

/*
 *@@ Cleanup:
 *      protected helper method to free
 *      allocated resources.
 *
 *      This is an internal method. Do not
 *      call this from user code.
 */

void BSString::Cleanup()
{
    xstrClear(&_str);
    ClearShiftTable();
}

/*
 *@@ ClearShiftTable:
 *      clears the shift table used for fast string
 *      searches.
 *
 *@@added V0.9.6 (2000-11-13) [umoeller]
 */

void BSString::ClearShiftTable()
{
    if (_pShiftTable)
    {
        free(_pShiftTable);
        _pShiftTable = 0;
    }
    _fRepeat = FALSE;
}

/*
 *@@ CopyFrom:
 *      protected helper method to set the member string
 *      from some other string.
 *
 *      This does not clear the member memory. Call
 *      Cleanup() before if memory has been allocated.
 *      However, this guarantees to set all member
 *      fields correctly.
 *
 *      This can safely be called from the copy constructors
 *      because we have not allocated any memory then.
 *
 *      However, this also gets called from the assign()
 *      methods and the "=" operators... we must call
 *      Cleanup() before then.
 *
 *      This is an internal method. Do not call this from
 *      user code.
 */

void BSString::CopyFrom(const char *psz)
{
    xstrInitCopy(&_str, psz, 0);

    _pShiftTable = 0;
    _fRepeat = FALSE;
}

/*
 *@@ CopyFrom:
 *      second CopyFrom to copy from another BSString.
 *
 *@@changed V0.9.7 (2001-01-07) [umoeller]: added n param
 */

void BSString::CopyFrom(const BSString &s,
                        size_type ulPos,        // in: copy offset; defaults to 0
                        size_type n)            // in: no. of chars to copy, defaults to npos
{
    // first clear all fields, or otherwise xstrcpy goes boom
    memset(&_str, 0, sizeof(XSTRING));

    if (ulPos < s._str.ulLength)
    {
        // start is valid:
        // calc start address
        const char  *pcszSource = s._str.psz + ulPos;

        // calc no. of chars to copy
        ULONG       cCopy;
        if (n == npos)
            // copy all:
            cCopy = s._str.ulLength - ulPos;
        else
        {
            cCopy = n;
            if (cCopy > (s._str.ulLength - ulPos))
                cCopy = s._str.ulLength - ulPos;
        }

        xstrcpy(&_str,
                pcszSource,
                cCopy);

        // if source string has a shift table and we copy
        // from the beginning, copy that shift table too
        if (    (s._pShiftTable != NULL)
             && (ulPos == 0)
             && (n == npos)
           )
        {
            _pShiftTable = (size_t*)malloc(sizeof(size_t) * 256);
            memcpy(_pShiftTable, s._pShiftTable, sizeof(size_t) * 256);
            _fRepeat = s._fRepeat;
        }
        else
        {
            // no shift table:
            _pShiftTable = 0;
            _fRepeat = FALSE;
        }
    }
    else
    {
        _pShiftTable = 0;
        _fRepeat = FALSE;
    }
}

/*
 *@@ BSString:
 *      default constructor to create an empty string.
 */

BSString::BSString() : BSRoot(tBSString)
{
    Init();
}

/*
 *@@ BSString:
 *      destructor.
 */

BSString::~BSString()
{
    Cleanup();
}

/*
 *@@ BSString:
 *      first (default) copy constructor.
 *
 *      This gets called when BSStrings
 *      are thrown as exception information!
 *
 *@@added V0.9.1 (2000-01-07) [umoeller]
 */

BSString::BSString(const BSString &s) : BSRoot(tBSString)
{
    CopyFrom(s);
}

/*
 *@@ BSString:
 *      second copy constructor to copy from
 *      another BSString. As opposed to the
 *      first version, you can specify a substring
 *      here.
 *
 *      Characters are copied from s starting
 *      at ulPos. Copying is stopped at the end
 *      of s or if n characters have been copied.
 */

BSString::BSString(const BSString &s,       // in: source string
                   size_type ulPos,         // in: first pos to copy
                   size_type n)             // in: maximum no. of chars to copy, defaults to npos
    : BSRoot(tBSString)
{
    CopyFrom(s, ulPos, n);
}

/*
 *@@ BSString:
 *      third copy constructor to copy from a PSZ.
 */

BSString::BSString(const char *psz)         // in: string to copy
    : BSRoot(tBSString)
{
    CopyFrom(psz);
}

/*
 *@@ BSString:
 *      fourth copy constructor to extract a new
 *      BSString from an existing C string, which
 *      starts at p1 and ends at p2. p2 is _not_
 *      included.
 *
 *      Note: as far as I know, something like this
 *      is NOT defined with the C++ string class.
 */

BSString::BSString(const char *p1, const char *p2)
    : BSRoot(tBSString)
{
    _str.psz = strhSubstr(p1, p2);
    if (_str.psz)
    {
        _str.ulLength = p2 - p1;
        _str.cbAllocated = _str.ulLength + 1;
    }
    else
    {
        _str.ulLength = 0;
        _str.cbAllocated = 0;
    }

    _str.ulDelta = _str.ulLength * 10 / 100;
            // V0.9.9 (2001-03-09) [umoeller]

    _pShiftTable = 0;
    _fRepeat = 0;
}

/*
 *@@ assign:
 *
 *@@added V0.9.7 (2001-01-07) [umoeller]
 */

BSString& BSString::assign(const BSString &s)
{
    // avoid self assignment
    if (_str.psz != s._str.psz)
    {
        Cleanup();
        CopyFrom(s);
    }

    return (*this);
}

/*
 *@@ assign:
 *
 *@@added V0.9.7 (2001-01-07) [umoeller]
 */

BSString& BSString::assign(const BSString &s,
                           size_type ulPos,     // in: start copy, defaults to 0
                           size_type n)         // in: copy count, defaults to npos
{
    // avoid self assignment
    if (_str.psz != s._str.psz)
    {
        Cleanup();
        CopyFrom(s, ulPos, n);

        // else: all fields are zero from Cleanup()
    }

    return (*this);
}

/*
 *@@ operator=:
 *      first operator= to copy from another BSString.
 */

BSString& BSString::operator=(const BSString &s) // in: string to copy
{
    return (assign(s));
}

/*
 *@@ operator=:
 *      second operator= to copy from a PSZ.
 *
 *@@changed V0.9.3 (2000-04-30) [umoeller]: fixed memory leak
 */

BSString& BSString::operator=(const char *psz) // in: string to copy
{
    // avoid self-assignment
    if (_str.psz != psz)
    {
        Cleanup();
        CopyFrom(psz);
    }
    return (*this);
}

/*
 *@@ operator=:
 *      third operator= to copy from a single character.
 *
 *@@changed V0.9.3 (2000-04-30) [umoeller]: fixed memory leak
 *@@changed V0.9.4 (2000-07-28) [umoeller]: added return *this
 */

BSString& BSString::operator=(char c) // in: character to copy
{
    Cleanup();
    char sz[2];
    sz[0] = c;
    sz[1] = 0;
    CopyFrom(sz);
    return (*this);
}

/*
 *@@ reserve:
 *      reserves additional memory for the string.
 *
 *      stExtra specifies the minimum byte count that should
 *      be allocated for the string. If stExtra is less than
 *      the currently allocated bytes, nothing happens. Otherwise
 *      the string is reallocated immediatly.
 *
 *      Useful for preparing a string for multiple subsequent
 *      operations which are likely to cause reallocations, e.g.
 *      string replacements. Call this method before those
 *      operations to make sure that the string does not get
 *      reallocated with every operation.
 *
 *@@added V0.9.7 (2001-01-07) [umoeller]
 */

void BSString::reserve(size_type stExtra) // in: extra bytes to reserve, defaults to 0
{
    if (stExtra)
    {
        xstrReserve(&_str, stExtra);
    }
}

/*
 *@@ operator[]:
 *      returns the ul'th character of the
 *      string or 0 if ul is too large.
 */

char BSString::operator[](const size_type ul)       // in: character offset
               const
{
    if (    (_str.psz)
         && (ul < _str.ulLength)
       )
        return (_str.psz[ul]);

    return ('\0');
}

/*
 *@@ compare:
 *      returns 0 if the two strings are the
 *      same, -1 if "this" is smaller, 1 if
 *      "s" is smaller.
 */

int BSString::compare(const BSString &s)
              const
{
    size_type ul1 = size(),
              ul2 = s.size();

    if ((ul1) && (ul2))
        return (strcmp(_str.psz, s._str.psz));
    else if (ul1)
        // ul1 defined, but ul2 not: string 2 is smaller
        return (1);
    else if (ul2)
        // ul1 not defined, but ul2 is: this is smaller
        return (-1);
    else
        // none defined:
        return (0);
}

/*
 *@@ compare:
 *      returns 0 if the two strings are the
 *      same, -1 if "this" is smaller, 1 if
 *      "s" is smaller.
 *
 *      As opposed to the standard compare,
 *      this only compares the first "n"
 *      characters of the member string,
 *      counting from ulPos.
 */

int BSString::compare(size_type ulPos,
                      size_type n,
                      const BSString &s)
              const
{
    BSString strTemp(*this, ulPos, n);
    return (strTemp.compare(s));
}

/*
 *@@ append:
 *      appends the given BSString to this.
 *      Returns *this.
 *
 *@@changed V0.9.7 (2001-01-15) [umoeller]: ClearShiftTable was missing
 */

BSString& BSString::append(const BSString &s)       // in: string to append
{
    xstrcat(&_str, s._str.psz, s._str.ulLength);
    ClearShiftTable();      // V0.9.7 (2001-01-15) [umoeller]
    return (*this);
}

/*
 *@@ append:
 *      appends the given PSZ to this.
 *      Returns *this.
 */

BSString& BSString::append(const char *psz)         // in: string to append
{
    xstrcat(&_str, psz, 0);
    ClearShiftTable();
    return (*this);
}

/*
 *@@ append:
 *      appends the given single character to this.
 *      Returns this.
 *
 *@@changed V0.9.7 (2001-01-15) [umoeller]: now using xstrcatc
 */

BSString& BSString::append(char c)            // in: character to append
{
    xstrcatc(&_str, c);
    ClearShiftTable();
    return (*this);
}

/*
 *@@ erase:
 *      removes at max n characters from the
 *      string, starting at ulPos.
 *
 *      Always returns *this.
 */

BSString& BSString::erase(size_type ulPos,      // in: start pos, defaults to 0
                          size_type n)          // in: count, defaults to npos
{
    size_type Size = size();

    if (ulPos < Size)
    {
        if (n == npos)
        {
            // delete rest of string: that's easy
            _str.psz[ulPos] = '\0';
            _str.ulLength = ulPos;
                // allocation doesn't change
        }
        else if (ulPos + n < Size)
        {
            // overwrite the part of the string at ulPos
            // with the following part, starting at ulPos + n
            char *pSource = _str.psz + ulPos + n,
                 *pTarget = _str.psz + ulPos;
            strcpy(pTarget, pSource);
            _str.ulLength = strlen(_str.psz);     // could be optimized...
                // allocation doesn't change
        }

        ClearShiftTable();
    }

    return *this;
}

/*
 *@@ find:
 *      first version of find.
 *
 *      Finds the first position of a substring in the string.
 *
 *      The search is started at iPos, which defaults to 0.
 *      Returns BSString::npos if not found.
 */

size_type BSString::find(const BSString &strFind,   // in: string to find
                         size_type ulPos)           // in: offset to start at (defaults to 0)
                    const
{
    if (ulPos < size())
    {
        char *p = strstr(_str.psz + ulPos, strFind._str.psz);
        if (p)
            // found: calculate offset
            return (p - _str.psz);
    }

    return (npos);
}

/*
 *@@ find:
 *      second version of find.
 *
 *      Finds the first position of a substring in the string.
 *
 *      The search is started at iPos, which defaults to 0.
 *      Returns BSString::npos if not found.
 */

size_type BSString::find(const char *pszFind,   // in: string to find
                         size_type ulPos)       // in: offset to start at (defaults to 0)
                    const
{
    if (ulPos < size())
    {
        char *p = strstr(_str.psz + ulPos, pszFind);
        if (p)
            // found: calculate offset
            return (p - _str.psz);
    }

    return (npos);
}

/*
 *@@ find:
 *      third version of find.
 *
 *      Finds the first position of a character in the string.
 *      The search is started at iPos, which defaults to 0.
 *
 *      Returns BSString::npos if not found.
 */

size_type BSString::find(char c,            // in: character to find
                         size_type ulPos)   // in: offset to start at (defaults to 0)
                    const
{
    if (ulPos < size())
    {
        char *p = strchr(_str.psz + ulPos, c);
        if (p)
            // found: calculate offset
            return (p - _str.psz);
    }

    return (npos);
}

/*
 *@@ rfind:
 *      finds the last position of a character in the string.
 *
 *      Returns BSString::npos if not found.
 */

size_type BSString::rfind(char c,           // in: character to find
                          size_type ulPos)  // in: position to search from (from the right);
                                            // defaults to npos (means end of string)
                    const
{
    char *p;

    size_type Size = size();

    if (Size)
    {
        if (    (ulPos == npos)
             || (ulPos > Size)
           )
            p = _str.psz + Size - 1;
        else
            p = _str.psz + ulPos;

        // search backwards
        for (;
             p >= _str.psz;
             p--)
            if (*p == c)
                return (p - _str.psz);
    }

    return (npos);
}

/*
 *@@ find_first_of:
 *      finds the first character in the member
 *      string which is one of the chars in achChars.
 *
 *      The search is started at ulPos, which
 *      defaults to 0.
 *
 *      Returns BSString::npos if not found.
 *
 *@@added V0.9.7 (2001-01-07) [umoeller]
 */

size_type BSString::find_first_of(const char *achChars, // in: chars array to look for
                                  size_type ulPos)      // in: start of search (defaults to 0)
                    const
{
    size_type Size = size();

    if (ulPos < Size)
    {
        char *p;
        size_type ul;
        for (p = _str.psz + ulPos, ul = ulPos;
             ul < Size;
             p++, ul++)
        {
            if (strchr(achChars, *p))
                // *p is in achChars:
                return (ul);
        }
    }

    // not found:
    return (npos);
}

/*
 *@@ find_first_not_of:
 *      finds the first character in the member
 *      string which is different from "c".
 *
 *      The search is started at ulPos, which
 *      defaults to 0.
 *
 *      Returns BSString::npos if not found.
 */

size_type BSString::find_first_not_of(char c,             // in: character to ignore
                                      size_type ulPos)    // in: start of search (defaults to 0)
                    const
{
    size_type Size = size();

    if (ulPos < Size)
    {
        char *p;
        size_type ul;
        for (p = _str.psz + ulPos, ul = ulPos;
             ul < Size;
             p++, ul++)
        {
            if (*p != c)
                return (ul);
        }
    }

    // not found:
    return (npos);
}

/*
 *@@ find_first_not_of:
 *      finds the first character in the member
 *      string which is NOT one of the chars in achChars.
 *
 *      The search is started at ulPos, which
 *      defaults to 0.
 *
 *      Returns BSString::npos if not found.
 *
 *@@added V0.9.7 (2001-01-07) [umoeller]
 */

size_type BSString::find_first_not_of(const char *achChars, // in: characters to ignore
                                      size_type ulPos)    // in: start of search (defaults to 0)
          const
{
    size_type Size = size();

    if (ulPos < Size)
    {
        char *p;
        size_type ul;
        for (p = _str.psz + ulPos, ul = ulPos;
             ul < Size;
             p++, ul++)
        {
            if (!strchr(achChars, *p))
                // *p is NOT in achChars:
                return (ul);
        }
    }

    // not found:
    return (npos);
}

/*
 *@@ replace:
 *      replaces "nThis" characters, starting at
 *      position ulPosThis, in the member string
 *      with "strReplace".
 *
 *      Returns *this.
 *
 *@@added V0.9.7 (2001-01-15) [umoeller]
 */

BSString& BSString::replace(size_type ulPosThis,        // in: ofs of first char to replace
                            size_type nThis,            // in: char count to replace
                            const BSString &strReplace) // in: string to replace with
{
    if (nThis == npos)
        // I'm not sure if the C++ string allows setting nThis to npos...
        // but we'll just support it, it won't hurt
        nThis = _str.ulLength - ulPosThis;

    xstrrpl(&_str,
            ulPosThis,
            nThis,
            strReplace._str.psz,
            strReplace._str.ulLength);

    ClearShiftTable();

    return (*this);
}

/*
 *@@ substr:
 *      creates a new string containing parts of
 *      the member string.
 *
 *      Copying is started at offset ulPos, from
 *      which n characters are copied. If
 *      (n == BSString::npos), all remaining characters
 *      are copied.
 *
 *      This always returns a string, but it may be
 *      empty if invalid parameters are specified.
 */

BSString BSString::substr(size_type ulPos,      // in: start pos, defaults to 0
                          size_type n)          // in: no. of characters to copy,
                                                // defaults to npos (== up to rest of string)
                   const
{
    return (BSString(*this, ulPos, n));
}

/*
 *@@ _printf:
 *      this replaces the current string with
 *      a new string, which is formatted according
 *      to printf() rules.
 *
 *      Note: as far as I know, something like this
 *      is NOT defined with the C++ string class.
 */

size_type BSString::_printf(const char *pszFormatString,
                            ...)
{
    size_type ul = 0;
    Cleanup();

    char szBuf[5000];
    va_list arg_ptr;

    // allocate memory
    va_start(arg_ptr, pszFormatString);
    ul = vsprintf(szBuf, pszFormatString, arg_ptr);
    va_end(arg_ptr);

    CopyFrom(szBuf);

    return (ul);
}

/*
 *@@ _find_word:
 *      searches for strWord in the string, whose offset
 *      is returned if found (or BSString::npos if not).
 *
 *      As opposed to BSString::find, this finds strWord
 *      only if it is a "word". A search string is
 *      considered a word if the character _before_
 *      it is in pcszBeginChars and the char _after_
 *      it is in pcszEndChars.
 *
 *      Example:
 +          BSString str("This is an example.");
 +          str._find_word(, "is");
 *
 *      returns the offset of "is", but not the "is" in "This".
 *
 *      Note: as far as I know, something like this
 *      is NOT defined with the C++ string class.
 */

size_type BSString::_find_word(BSString &strFind,
                               size_type ulPos,     // defaults to 0
                               const char *pcszBeginChars, //  = "\x0d\x0a ()/\\-,.",
                               const char *pcszEndChars) // = "\x0d\x0a ()/\\-,.:;")
                    const
{
    if (_str.psz)
    {
        if (ulPos < size())
        {
            // if this is the first time strFind is used as
            // a "find" string, allocate a shift table...
            if (strFind._pShiftTable == NULL)
            {
                strFind._pShiftTable = (size_t*)malloc(sizeof(size_t) * 256);
                strFind._fRepeat = FALSE;
            }

            char *p = xstrFindWord(&_str,
                                   ulPos,
                                   &strFind._str,
                                   strFind._pShiftTable,
                                   (PBOOL)&strFind._fRepeat,
                                   pcszBeginChars,
                                   pcszEndChars);
            if (p)
                // found: calculate offset
                return (p - _str.psz);
        }
    }

    return (npos);
}

/*
 *@@ _find_word:
 *      second version of _find_word, which takes
 *      a const PSZ as input.
 *
 *      Note: as far as I know, something like this
 *      is NOT defined with the C++ string class.
 */

size_type BSString::_find_word(const char *pszFind,
                               size_type ulPos,     // defaults to 0
                               const char *pcszBeginChars, //  = "\x0d\x0a ()/\\-,.",
                               const char *pcszEndChars) // = "\x0d\x0a ()/\\-,.:;")
                    const
{
    if (_str.psz)
    {
        if (ulPos < size())
        {
            char *p = strhFindWord(_str.psz + ulPos,
                                   pszFind,
                                   pcszBeginChars,
                                   pcszEndChars);
            if (p)
                // found: calculate offset
                return (p - _str.psz);
        }
    }

    return (npos);
}

/*
 *@@ _find_replace:
 *      replaces the first occurence of strFind
 *      with strReplaceWith which is found after
 *      the ulStartPos offset.
 *
 *      Returns the offset at which strFind was
 *      found or BSString::npos if it wasn't found.
 *
 *      To replace all occurences in the string,
 *      keep calling this function until BSString::npos
 *      is returned, like this:
 *
 +          string str = "part 1, part 2";
 +          ULONG ulPos = 0;
 +          while (     (ulPos == str._find_replace("part", "piece", ulPos))
 +                   != string::npos)
 +              ;
 *
 *      Note: as far as I know, something like this
 *      is NOT defined with the C++ string class.
 */

size_type BSString::_find_replace(BSString &strFind,
                                  const BSString &strReplaceWith,
                                  size_type *pulPos) // in: where to start search (defaults to 0)
{
    // if this is the first time strFind is used as
    // a "find" string, allocate a shift table...
    if (strFind._pShiftTable == NULL)
    {
        strFind._pShiftTable = (size_t*)malloc(sizeof(size_t) * 256);
        strFind._fRepeat = FALSE;
    }

    if (    (strFind._pShiftTable)
         && (xstrFindReplace(&_str,
                             pulPos,
                             &strFind._str,
                             &strReplaceWith._str,
                             strFind._pShiftTable,
                             (PBOOL)&strFind._fRepeat))
       )
    {
        // found:
        // we've changed, so invalidate our own shift table
        ClearShiftTable();
        return (*pulPos);
    }

    // unchanged:
    return (npos);
}

/*
 *@@ _find_replace:
 *      second version of _find_replace, which takes
 *      a const PSZ as input.
 *
 *      Note: as far as I know, something like this
 *      is NOT defined with the C++ string class.
 */

size_type BSString::_find_replace(const char *pszFind,
                                  const BSString &strReplaceWith,
                                  size_type *pulPos)
{
    // create temporary "find" string
    BSString strFind(pszFind);
    // initialize shift table
    strFind._pShiftTable = (size_t*)malloc(sizeof(size_t) * 256);
    strFind._fRepeat = FALSE;
    return (_find_replace(strFind,
                          strReplaceWith,
                          pulPos));
}

/*
 *@@ _find_replace:
 *      third version of _find_replace, which simply
 *      replaces the first occurence of one character
 *      with another.
 *
 *      This is very inexpensive because no memory
 *      management is involved.
 *
 *      Note: as far as I know, something like this
 *      is NOT defined with the C++ string class.
 *
 *@@added V0.9.6 (2000-11-01) [umoeller]
 */

size_type BSString::_find_replace(char cFind,
                                  char cReplace,
                                  size_type *pulPos)
{
    PSZ p = strchr(_str.psz, cFind);
    if (p)
    {
        *p = cReplace;
        // we've changed, so invalidate our own shift table
        ClearShiftTable();
        return (p - _str.psz);
    }

    return (npos);
}

/*
 *@@ _take_from:
 *      takes the specified XSTRING for self
 *      without allocating a new buffer. Useful
 *      for conversion of XSTRING's to BSString's.
 *
 *@@added V0.9.16 (2001-10-11) [umoeller]
 */

int BSString::_take_from(XSTRING &str)
{
    Cleanup();
    memcpy(&_str, &str, sizeof(XSTRING));

    return (_str.ulLength);
}

/*
 *@@ _extract_from:
 *      sets the string to what strhSubstr
 *      returns.
 */

int BSString::_extract_from(const char *pStart,
                            const char *pEnd)
{
    Cleanup();
    _str.psz = strhSubstr(pStart, pEnd);
    if (_str.psz)
    {
        _str.ulLength = pEnd - pStart;
        _str.cbAllocated = _str.ulLength + 1;
    }
    // else: all fields are zero from Cleanup()

    return (_str.ulLength);
}

/*
 *@@ _extract_word:
 *      extracts a word from the member string,
 *      which is written into strTarget.
 *
 *      Returns 1 (TRUE) if the specified word still
 *      existed or 0 otherwise.
 *
 *      Note: as far as I know, something like this
 *      is NOT defined with the C++ string class.
 *
 *@@added V0.9.7 (2001-01-07) [umoeller]
 */

int BSString::_extract_word(unsigned long ulIndex,  // in: word index
                            BSString &strTarget,    // out: word
                            size_type ulPos,        // defaults to 0
                            const char *pcszBeginChars, //  = "\x0d\x0a ()/\\-,.",
                            const char *pcszEndChars) // = "\x0d\x0a ()/\\-,.:;")
    const
{
    int rc = 0;

    unsigned long ul = 0;
    for (ul = 0;
         ul <= ulIndex;
         ul++)
    {
        ULONG ulStart = find_first_not_of(pcszBeginChars, ulPos);
        if (ulStart == npos)
            break;
        else
        {
            ULONG ulEnd = find_first_of(pcszEndChars, ulStart + 1);
            if (ulEnd == npos)
            {
                if (ul == ulIndex)
                {
                    // copy till the end
                    strTarget.assign(*this,
                                     ulStart);
                    rc = 1;
                    break;
                }
                else
                    // not there yet: get outta here
                    break;
            }
            else
                if (ul == ulIndex)
                {
                    // that's the word we want:
                    strTarget.assign(*this,
                                     ulStart,
                                     (ulEnd - ulStart));
                    rc = 1;
                    break;
                }
                else
                    // we need more words:
                    ulPos = ulEnd;
                    // and search on
        }
    }

    return (rc);
}

/*
 *@@ _format:
 *      this removes all EOL's and double spaces.
 *      Also removes leading and trailing spaces.
 *
 *      Note: as far as I know, something like this
 *      is NOT defined with the C++ string class.
 */

VOID BSString::_format()
{
    if (_str.psz)
    {
        PSZ p;
        // replace \n, \r with spaces (length is constant)
        for (p = _str.psz;
             p < _str.psz + _str.ulLength;
             p++)
        {
            if (*p == 0x0a)
                *p = ' ';
            else if (*p == 0x0d)
                *p = ' ';
        }

        // remove double spaces
        for (p = _str.psz;
             p < _str.psz + _str.ulLength;
             p++)
        {
            if ((*p == ' ') && (*(p+1) == ' '))
            {
                PSZ p2 = p;
                while (*p2)
                {
                    *p2 = *(p2+1);
                    p2++;
                }
                _str.ulLength--;
                p--;
            }
        }

        // remove leading spaces
        for (p = _str.psz;
             p < _str.psz + _str.ulLength;
             p++)
        {
            if (*p == ' ')
            {
                PSZ p2 = p;
                while (*p2)
                {
                    *p2 = *(p2+1);
                    p2++;
                }
                _str.ulLength--;
            }
            else
                break;
        }

        // remove trailing spaces
        for (p = _str.psz + _str.ulLength - 1;
             p > _str.psz;
             p--)
        {
            if (*p == ' ')
            {
                *p = 0;
                _str.ulLength--;
            }
            else
                break;
        }

        // we've changed, so invalidate our own shift table
        ClearShiftTable();
    }
}

int operator==(const BSString &s1, const BSString &s2)
{
    return (s1.compare(s2) == 0);
}

int operator==(const char *psz1, const BSString &s2)
{
    return (s2.compare(psz1) == 0);
}

int operator==(const BSString &s1, const char *psz2)
{
    return (s1.compare(psz2) == 0);
}

int operator!=(const BSString &s1, const BSString &s2)
{
    return (s1.compare(s2) != 0);
}

int operator!=(const char *psz1, const BSString &s2)
{
    return (s2.compare(psz1) != 0);
}

int operator!=(const BSString &s1, const char *psz2)
{
    return (s1.compare(psz2) != 0);
}

int operator<(const BSString &s1, const BSString &s2)
{
    return (s1.compare(s2) < 0);
}

BSString operator+(const BSString &s1, const BSString &s2)
{
    BSString str(s1);
    str.append(s2);
    return (str);
}

BSString operator+(const char *psz1, const BSString &s2)
{
    BSString str(psz1);
    str.append(s2);
    return (str);
}

BSString operator+(const BSString &s1, const char *psz2)
{
    BSString str(s1);
    str.append(psz2);
    return (str);
}


// test case

/* int main()
{
    BSString str1("This is a test string."),
             str2(str1, 0);

    printf("Original was: %s\n", str1.c_str());
    printf("Second is:    %s\n", str2.c_str());

    str2 = str1 + " Now we appended something.";
    printf("Third is:     %s\n", str2.c_str());

    BSString strFind = "e";
    BSString strRepl = "DJASDKLASDPO(u";
    ULONG ulPos = 0;
    while (str2._find_replace(strFind, strRepl, &ulPos)
                != string::npos)
            ;

    printf("After repl:   %s\n", str2.c_str());

    return (0);
}
*/
