
/*
 * bs_string.h:
 *      header file for bs_string.cpp. See remarks there.
 *
 *      Warning: if you include this header, you'll get a
 *      typedef for "string" to BSString.
 *
 *@@include #include "helpers\stringh.h"
 *@@include #include "helpers\xstring.h"
 *@@include #include "base\bs_string.h"
 */

/*
 *      This file Copyright (C) 1999-2000 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#ifndef WARPIN_STRING_HEADER_INCLUDED
    #define WARPIN_STRING_HEADER_INCLUDED

    // STL support
    // #include <map>

    typedef unsigned long size_type;

    /*
     *@@ BSString:
     *      simple string class vaguely modelled after the C++ ANSI
     *      string class which IBM VAC 3.08 has no support for.
     *
     *      This implements the "BSString" class and typedefs "string"
     *      to "BSString". As a result, you can use either "BSString"
     *      or "string".
     *
     *      <B>Introduction</B>
     *
     *      If you're not familiar with the C++ "string" class, here's
     *      a short introduction.
     *
     *      Basically, the "string" class allows you to use strings
     *      as a basic type, like int, char, long, and so on, without
     *      having to worry about memory management.
     *
     *      Here is an example:
     +
     +          string str = "Hello."
     +
     +          if (str == "Hello")
     +          {
     +              string str = "We should say ";
     +              string str2 = str + "Hello now."
     +
     +              printf("%s", str2.c_str());
     +          }
     *
     *      This shows some of the most convenient features of the
     *      string class:
     *
     *      -- You can just create an instance of it on the stack.
     *         The destructor will automatically free the memory
     *         that has been allocated for the string -- no more
     *         memory leaks as with the C functions.
     *
     *      -- You can compare strings using ==, !=, <, >. (This
     *         calls strcmp internally.)
     *
     *      -- You can concatenate strings using the "+" operator.
     *
     *      -- If you need a C-type string (e.g. for OS/2 API
     *         functions), use the BSString::c_str() method,
     *         which returns a "const char*" pointer to the string
     *         in the class.
     *
     *      In addition, the string class has methods for more
     *      complicated things, such BSString::find and
     *      BSString::replace.
     *
     *      <B>C++ ANSI "string" class compatibility</B>
     *
     *      Unless marked otherwise, all methods and operators are
     *      compatible with the standard "string" class.
     *
     *      However, we only support single-byte (8-bit) character
     *      sets at this point, and only a subset of the "string"
     *      methods and operators.
     *
     *      Non-standard functions are marked with a leading
     *      underscore (e.g. BSString::_printf).
     *
     *      Also be warned, this string class is more expensive
     *      than the C++ version because we can NOT do reference counting
     *      on memory buffers. In other words, each instance has its own
     *      copy of the string, even if two strings are equal. This is
     *      most obvious when one string is created from another; you
     *      then have two identical copies.
     *
     *      With V0.9.6, I have made improvements on memory management
     *      though. Memory no longer gets reallocated with every change,
     *      but only if necessary.
     *
     *      Finally, this class is NOT thread-safe. If methods are
     *      operating on the same instance on several threads at
     *      the same time, this will DEFINITELY crash.
     */

    class BSString : public BSRoot
    {
        public:
            DECLARE_CLASS(BSString);

        #ifdef __IBMCPP__
            friend class BSString;
        #endif
        // 2000-01-15: This line is extraordinary useless - class instances
        //             of the same type are always friends (czw)
        // V0.9.1 (2000-02-01) [umoeller]: Hi Jens. Not with VAC++, apparently,
        //             because without this line, this wouldn't compile.

        protected:
            XSTRING         _str;

            size_t          *_pShiftTable;
            int             _fRepeat;

            void Init();
            void Cleanup();
            void ClearShiftTable();

            void CopyFrom(const char *psz);
            void CopyFrom(const BSString &s,
                          size_type ulPos = 0,
                          size_type n = npos);

        public:
            static const size_type npos;

            BSString();
            ~BSString();

            // default copy constructor;
            // required for exception handling!
            BSString(const BSString &s);

            BSString(const BSString &s,
                     size_type ulPos,
                     size_type n = npos);
            BSString(const char *psz);
            BSString(const char *p1, const char *p2);

            BSString& assign(const BSString &s);
            BSString& assign(const BSString &s,
                             size_type ulPos,
                             size_type n = npos);

            BSString& operator=(const BSString &s);
            BSString& operator=(const char *psz);
            BSString& operator=(char c);

            /*
             *@@ c_str:
             *      returns the string member as a
             *      classic const C string. Note that
             *      this _never_ returns NULL. If the
             *      member string is empty, this
             *      returns a pointer to a null byte.
             */

            const char* c_str()
                        const
            {
                if (_str.psz)
                    return (_str.psz);
                else
                    return "";
            }

            /*
             *@@ size:
             *      returns the no. of characters in
             *      the string (excluding our internal
             *      null terminator) or 0 if the string
             *      is empty.
             */

            inline size_type size()
                             const
            {
                return (_str.ulLength);
            }

            /*
             *@@ length:
             *      STL string compatibility function,
             *      same as BSString::size().
             */

            inline size_type length()
                             const
            {
                return (_str.ulLength);
            }

            /*
             *@@ capacity:
             *      returns the amount of memory presently
             *      allocated for the string. This is at
             *      least what size() or length() return.
             *
             *@@added V0.9.6 (2000-10-31) [umoeller]
             */

            inline size_type capacity()
                             const
            {
                return (_str.cbAllocated);
            }

            void reserve(size_type stExtra = 0);

            /*
             *@@ operator():
             *      returns 1 (TRUE) if the string
             *      contains something, 0 (FALSE)
             *      otherwise.
             *
             *      Note: as far as I know, this operator
             *      is NOT defined with the C++ string class.
             */

            inline int operator()()
                       const
            {
                return (size() != 0);
            }

            /*
             *@@ empty:
             *      returns 1 (TRUE) if the string
             *      contains nothing, 0 (FALSE) otherwise.
             */

            inline int empty()
                       const
            {
                return (size() == 0);
            }

            char operator[](const size_type ul) const;

            int compare(const BSString &s) const;
            int compare(size_type ulPos, size_type n, const BSString &s) const;

            BSString& append(const BSString &s);
            BSString& append(const char *psz);
            BSString& append(char c);

            /*
             *@@ operator+=:
             *      shortcut to append(BSString &s).
             *
             *@@changed V0.9.2 (2000-03-29) [umoeller]: added return *this
             */

            BSString& operator+=(const BSString &s)     // in: string to append
            {
                append(s);
                return *this;
            }

            /*
             *@@ operator+=:
             *      shortcut to append(const char *psz).
             *
             *@@changed V0.9.2 (2000-03-29) [umoeller]: added return *this
             */

            BSString& operator+=(const char *psz)     // in: string to append
            {
                append(psz);
                return *this;
            }

            /*
             *@@ operator+=:
             *      shortcut to append(char c).
             *
             *@@changed V0.9.2 (2000-03-29) [umoeller]: added return *this
             */

            BSString& operator+=(char c)     // in: character to append
            {
                append(c);
                return *this;
            }

            BSString& erase(size_type ulPos = 0, size_type n = npos);

            size_type find(const BSString &strFind, size_type ulPos = 0) const;
            size_type find(const char *pszFind, size_type ulPos = 0) const;
            size_type find(char c, size_type ulPos = 0) const;

            size_type rfind(char c, size_type ulPos = npos) const;

            size_type find_first_of(const char *achChars, size_type ulPos = 0) const;

            size_type find_first_not_of(char c, size_type ulPos = 0) const;
            size_type find_first_not_of(const char *achChars, size_type ulPos = 0) const;

            BSString& replace(size_type ulPosThis,
                              size_type nThis,
                              const BSString &strReplace);

            /* BSString& replace(size_type ulPosThis,
                              size_type nThis,
                              const BSString &strReplace,
                              size_type ulPosReplace = 0,
                              size_type nReplace = npos);

            BSString& replace(size_type ulPosThis,
                              size_type nThis,
                              const char *pszReplace,
                              size_type nReplace);

            BSString& replace(size_type ulPosThis,
                              size_type nThis,
                              const char *pszReplace); */

            BSString substr(size_type ulPos = 0,
                            size_type n = npos)
                     const;

            /*
             *@@ swap:
             *      swaps the contents of this instance
             *      with that of another BSString.
             *
             *      This is very fast because only the
             *      pointers are exchanged.
             *
             *@@added V0.9.6 (2000-10-31) [umoeller]
             */

            void swap(BSString &str)
            {
                XSTRING strTemp;
                memcpy(&strTemp, &str._str, sizeof(XSTRING));
                memcpy(&str._str, &_str, sizeof(XSTRING));
                memcpy(&_str, &strTemp, sizeof(XSTRING));
                // we've changed, so invalidate our own shift table
                ClearShiftTable();
                str.ClearShiftTable();
            }

            size_type _printf(const char *pszFormatString, ...);

            size_type _find_word(BSString &strFind,
                                 size_type ulPos = 0,
                                 const char *pcszBeginChars = "\x0d\x0a ()/\\-,.;*",
                                 const char *pcszEndChars   = "\x0d\x0a ()/\\-,.:;*&'")
                      const;

            size_type _find_word(const char *pszFind,
                                 size_type ulPos = 0,
                                 const char *pcszBeginChars = "\x0d\x0a ()/\\-,.;*",
                                 const char *pcszEndChars   = "\x0d\x0a ()/\\-,.:;*&'")
                      const;

            size_type _find_replace(BSString &strFind,
                                    const BSString &strReplaceWith,
                                    size_type *pulPos);
            size_type _find_replace(const char *pszFind,
                                    const BSString &strReplaceWith,
                                    size_type *pulPos);
            size_type _find_replace(char cFind,
                                    char cReplace,
                                    size_type *pulPos);

            int _take_from(XSTRING &str);

            int _extract_from(const char *pStart,
                              const char *pEnd);

            int _extract_word(unsigned long ulIndex,
                              BSString &strTarget,
                              size_type ulPos = 0,
                              const char *pcszBeginChars = "\x0d\x0a ()/\\-,.;*",
                              const char *pcszEndChars   = "\x0d\x0a ()/\\-,.:;*&'")
                const;

            void _format();
    };

    int operator==(const BSString &s1, const BSString &s2);

    int operator==(const char *psz1, const BSString &s2);

    int operator==(const BSString &s1, const char *psz2);

    int operator!=(const BSString &s1, const BSString &s2);

    int operator!=(const char *psz1, const BSString &s2);

    int operator!=(const BSString &s1, const char *psz2);

    int operator<(const BSString &s1, const BSString &s2);

    BSString operator+(const BSString &s1, const BSString &s2);

    BSString operator+(const char *psz1, const BSString &s2);

    BSString operator+(const BSString &s1, const char *psz2);

    /*
     *@@ string:
     *      typedef for using BSStrings like the
     *      standard string class.
     */

    typedef BSString string;

    /*
     *@@ BSStringMap:
     *      typedef for a BSString map.
     *
     *@@added V0.9.2 (2000-03-11) [umoeller]
     */

    /* typedef map< BSString,      // key class
                 BSString,      // data class
                 less<BSString>
               > BSStringMap; */

#endif

